"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const index_1 = require("../index");
describe('result types compatability', () => {
    it('not allow to assign results with not different error type to each other', () => {
        expect.assertions(2);
        const r1 = (0, index_1.fail)('e1');
        let r2 = (0, index_1.fail)('e2');
        expect(r1.err().type).toStrictEqual('e1');
        if (Math.random()) {
            r2 = r1;
        }
        expect(r1.err().type).toStrictEqual('e1');
    });
});
describe('result and success/failure types compatability', () => {
    it('success assignable to result', () => {
        expect.assertions(2);
        const r1 = (0, index_1.ok)('ok1');
        expect(r1.ok()).toStrictEqual('ok1');
        const r2 = {
            status: 'success',
            data: 'ok2',
        };
        expect(r2.data).toStrictEqual('ok2');
    });
    it('failure assignable to result', () => {
        expect.assertions(2);
        const r1 = (0, index_1.fail)('e1');
        expect(r1.err().type).toStrictEqual('e1');
        const r2 = {
            status: 'error',
            error: { type: 'e1' },
        };
        expect(r2.error.type).toStrictEqual('e1');
    });
    it('result is assignable to result with never as success or error', () => {
        expect.assertions(2);
        const r1 = { status: 'success', data: 'ok1' };
        expect(r1.status === 'success' ? r1.data : undefined).toStrictEqual('ok1');
        const r2 = {
            status: 'error',
            error: { type: 'e1' },
        };
        expect(r2.status === 'error' ? r2.error.type : undefined).toStrictEqual('e1');
    });
    it('narrowing success type', () => {
        expect.assertions(0);
        const r1 = (0, index_1.ok)('ok1');
        if (r1.isOk()) {
        }
        if (r1.isErr()) {
        }
    });
    it('narrowing error type', () => {
        expect.assertions(0);
        const r1 = (0, index_1.fail)('e1');
        if (r1.isErr()) {
        }
        if (r1.isOk()) {
        }
    });
    it('narrowing result to error or success type', () => {
        expect.assertions(0);
        const r1 = Math.random() ? (0, index_1.ok)('1') : (0, index_1.fail)('e1');
        if (r1.isOk()) {
        }
        if (r1.isErr()) {
        }
    });
    it('combines two results into one', () => {
        expect.assertions(2);
        const r1 = Math.random() ? (0, index_1.ok)({ r1: '1' }) : (0, index_1.fail)('e2');
        const r2 = Math.random() ? (0, index_1.ok)({ r2: 2 }) : (0, index_1.fail)('e3');
        let r3;
        r3 = (0, index_1.ok)({
            ra: r1.ok(),
            rb: r2.ok(),
        });
        expect(r3.ok()).toMatchObject({ ra: { r1: '1' }, rb: { r2: 2 } });
        r3 = (0, index_1.fail)('e1');
        expect(r3.err().type).toStrictEqual('e1');
    });
});
describe('result', () => {
    it('sets the code correctly', () => {
        expect.assertions(1);
        const result = (0, index_1.fail)('', { code: 400 });
        expect(result.code).toStrictEqual(400);
    });
    it('error is an exception', () => {
        expect.assertions(1);
        const result = (0, index_1.fail)(undefined);
        expect(result).toBeInstanceOf(Error);
    });
    it('sets the message correctly', () => {
        expect.assertions(2);
        const result = (0, index_1.fail)('e1', { message: 'Some error' });
        expect(result.message).toStrictEqual('Some error');
        expect(result.err().message).toStrictEqual('Some error');
    });
    it('sets the empty string message correctly', () => {
        expect.assertions(3);
        const result = (0, index_1.fail)('');
        expect(result.message).toStrictEqual('Unknown');
        expect(result.err().message).toBeUndefined();
        const result2 = (0, index_1.err)(undefined);
        expect(result2.message).toStrictEqual('Unknown');
    });
    it('sets the type message correctly', () => {
        expect.assertions(2);
        const result = (0, index_1.fail)('Err type');
        expect(result.message).toStrictEqual('Err type');
        expect(result.err().message).toBeUndefined();
    });
    it('returns data with ok method', () => {
        expect.assertions(4);
        const result = Math.random() !== -1 ? (0, index_1.ok)('data') : (0, index_1.fail)('');
        expect(result.isOk()).toBe(true);
        expect(result.isErr()).toBe(false);
        expect(result.ok()).toStrictEqual('data');
        expect(() => {
            return result.err();
        }).toThrow("Can't access error on data");
    });
    it('works when fail called with undefined', () => {
        expect.assertions(4);
        const result = Math.random() !== -1 ? (0, index_1.err)(undefined) : (0, index_1.ok)('');
        expect(result.isOk()).toBe(false);
        expect(result.isErr()).toBe(true);
        expect(() => {
            return result.ok();
        }).toThrow("Can't access data on error");
        expect(result.err()).toBeUndefined();
    });
    it('returns error with err method', () => {
        expect.assertions(5);
        const result = Math.random() !== -1 ? (0, index_1.fail)('e2', { stringAdded: 'e2data' }) : (0, index_1.ok)('');
        expect(result.isOk()).toBe(false);
        expect(result.isErr()).toBe(true);
        expect(() => {
            return result.ok();
        }).toThrow("Can't access data on error");
        expect(result.err().type).toStrictEqual('e2');
        expect(result.err().stringAdded).toStrictEqual('e2data');
    });
    it('will check errors exhaustively', () => {
        expect.assertions(1);
        const test1 = (r) => {
            const e = r.err();
            switch (e.type) {
                case 'e1':
                    return 1;
                case 'e2':
                    return 2;
                case 'e3':
                    return 3;
                default:
                    (0, index_1.nope)(e);
                    return 0;
            }
        };
        const e1 = (0, index_1.fail)('e1');
        expect(test1(e1)).toStrictEqual(1);
    });
    it('different options passed correctly', () => {
        expect.assertions(8);
        const e1 = (0, index_1.fail)('e1', { fatal: true, level: index_1.ErrLevel.Crit });
        expect(e1.err().fatal).toStrictEqual(true);
        expect(e1.err().notify).toBeUndefined();
        expect(e1.err().retry).toBeUndefined();
        expect(e1.err().level).toStrictEqual(index_1.ErrLevel.Crit);
        const e2 = (0, index_1.fail)('e1', { retry: true, notify: true });
        expect(e2.err().fatal).toBeUndefined();
        expect(e2.err().notify).toStrictEqual(true);
        expect(e2.err().retry).toStrictEqual(true);
        expect(e2.err().level).toBeUndefined();
    });
    it('type overwrite', () => {
        expect.assertions(1);
        const e1 = (0, index_1.fail)('test1', { type: 'test2' });
        expect(e1.err().type).toStrictEqual('test1');
    });
    it('works with errors groups', () => {
        expect.assertions(4);
        const err1 = (0, index_1.fail)('ErrsErr3', { addedString: 'str' });
        expect(err1.err().type).toStrictEqual('ErrsErr3');
        const err2 = (0, index_1.fail)('ErrsErr2');
        const check = (e) => {
            if (e.type === 'ErrsErr1') {
                return 1;
            }
            if (e.type === 'ErrsErr2') {
                return 2;
            }
            if (e.type === 'ErrsErr3') {
                return 3;
            }
            (0, index_1.nope)(e);
            return 0;
        };
        expect(check(err2.err())).toStrictEqual(2);
        const err3 = (0, index_1.fail)('Err5', { addedNumber: 5 });
        expect(err3.err().type).toStrictEqual('Err5');
        expect(err3.isErr() ? (err3.error.type === 'Err5' ? err3.error.addedNumber : undefined) : undefined).toStrictEqual(5);
    });
    describe('infer error based on its type', () => {
        it('type inferred from common result', () => {
            expect.assertions(4);
            let r1 = (0, index_1.ok)('1');
            expect.assertions(1);
            expect(r1.ok()).toStrictEqual('1');
            r1 = (0, index_1.fail)('e1');
            expect(r1.err().type).toStrictEqual('e1');
            r1 = (0, index_1.fail)('e2', { stringAdded: 'e2data' });
            expect.assertions(4);
            expect(r1.isErr() ? (r1.error.type === 'e2' ? r1.error.stringAdded : null) : null).toStrictEqual('e2data');
            r1 = (0, index_1.fail)('e3', { numberAdded: 100 });
            expect(r1.isErr() ? (r1.error.type === 'e3' ? r1.error.numberAdded : null) : null).toStrictEqual(100);
        });
        it('types inferred from different fails and oks', () => {
            expect.assertions(4);
            const r1 = Math.random() !== -1 ? (0, index_1.ok)(1) : (0, index_1.fail)('e1');
            expect(r1.ok()).toStrictEqual(1);
            const e1 = (0, index_1.fail)('e2', { stringAdded: 'e2data' });
            const r2 = Math.random() !== -1 ? e1 : r1;
            expect(r2.err().type).toStrictEqual('e2');
            const r3 = Math.random() !== -1 ? r2 : r1;
            expect(r3.isErr() ? (r3.error.type === 'e2' ? r3.error.stringAdded : undefined) : undefined).toStrictEqual('e2data');
            const res3 = r3;
            expect(res3.isErr()).toBeTruthy();
        });
    });
});
//# sourceMappingURL=index.js.map