"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.nope = exports.toResult = exports.compare = exports.throwFail = exports.fail = exports.err = exports.ok = exports.complete = exports.FailureException = void 0;
const checks_1 = require("./checks");
class FailureException extends Error {
    constructor(t, e, o, m, c) {
        super(m || t);
        Object.setPrototypeOf(this, FailureException.prototype);
        this.status = 'error';
        this.order = o;
        this.error = e;
        this.code = c;
        this.name = 'FailureException';
    }
}
exports.FailureException = FailureException;
const maybeFailToResult = (result, prevResult) => {
    if ((0, checks_1.isSuccessLike)(result)) {
        return result;
    }
    let failure = result;
    if (!(0, checks_1.isHaveStatus)(result)) {
        let exception;
        if (typeof result === 'string') {
            exception = new FailureException(result, { type: result }, prevResult.order, prevResult.message, prevResult.code);
        }
        else if ((0, checks_1.isErr)(result)) {
            exception = new FailureException(result.type, result, prevResult.order, prevResult.message, prevResult.code);
        }
        else {
            throw new Error("Can't convert to error");
        }
        failure = exception;
    }
    return failure;
};
const transform = (future) => {
    const onErr = (cb) => {
        return transform(future.then(({ current, prev }) => {
            const currentRes = maybeFailToResult(current, prev);
            if (currentRes.status === 'success') {
                return { current: currentRes, prev: currentRes };
            }
            const nextRes = cb(currentRes.error, currentRes);
            return Promise.resolve(nextRes).then((next) => {
                return {
                    current: next,
                    prev: currentRes,
                };
            });
        }));
    };
    const onOk = (cb) => {
        return transform(future.then(({ current, prev }) => {
            const currentRes = maybeFailToResult(current, prev);
            if (currentRes.status === 'error') {
                return { current: currentRes, prev: currentRes };
            }
            const nextRes = cb(currentRes.data, currentRes);
            return Promise.resolve(nextRes).then((next) => {
                return {
                    current: next,
                    prev: currentRes,
                };
            });
        }));
    };
    const helper = {
        onOk,
        onFail: onErr,
        onErr,
        async res() {
            const { current, prev } = await future;
            return (0, exports.complete)(maybeFailToResult(current, prev));
        },
    };
    return helper;
};
const complete = (partial) => {
    const propsDefs = {
        isOk: {
            enumerable: false,
        },
        isErr: {
            enumerable: false,
        },
        ok: {
            enumerable: false,
        },
        err: {
            enumerable: false,
        },
        onOk: {
            enumerable: false,
        },
        onFail: {
            enumerable: false,
        },
        onErr: {
            enumerable: false,
        },
        res: {
            enumerable: false,
        },
        toString: {
            enumerable: false,
        },
    };
    if (partial.order === undefined) {
        delete partial.order;
    }
    if (partial.code === undefined) {
        delete partial.code;
    }
    const result = partial;
    result.toString = () => {
        if (result.status === 'success') {
            return `Success<${JSON.stringify(result.data)}>`;
        }
        if (result.status === 'error') {
            return `Failure<${JSON.stringify(result.error)}>`;
        }
        throw new Error('Unknown status');
    };
    result.isOk = () => {
        return result.status === 'success';
    };
    result.isErr = () => {
        return result.status === 'error';
    };
    result.ok = () => {
        if (result.status === 'success') {
            return result.data;
        }
        throw new Error("Can't access data on error");
    };
    result.err = () => {
        if (result.status === 'error') {
            return result.error;
        }
        throw new Error("Can't access error on data");
    };
    result.res = (async () => {
        return result;
    });
    result.onOk = (cb) => {
        if (result.status === 'error') {
            return result;
        }
        const nextRes = cb(result.data, result);
        if ((0, checks_1.isPromise)(nextRes)) {
            const transformed = nextRes.then((current) => {
                return { current, prev: result };
            });
            return transform(transformed);
        }
        return (0, exports.complete)(nextRes);
    };
    result.onFail = (cb) => {
        if (result.status === 'success') {
            return result;
        }
        const nextRes = cb(result.error, result);
        if ((0, checks_1.isPromise)(nextRes)) {
            const transformed = nextRes.then((current) => {
                return { current, prev: result };
            });
            return transform(transformed);
        }
        return (0, exports.complete)(nextRes);
    };
    result.onErr = (cb) => {
        if (result.status === 'success') {
            return result;
        }
        const nextRes = cb(result.error, result);
        if ((0, checks_1.isPromise)(nextRes)) {
            const transformed = nextRes.then((current) => {
                return { current, prev: result };
            });
            return transform(transformed);
        }
        return (0, exports.complete)(maybeFailToResult(nextRes, result));
    };
    return Object.defineProperties(result, propsDefs);
};
exports.complete = complete;
const ok = (data, { code, order, skip } = {}) => {
    if (skip) {
        order = -Infinity;
    }
    const partial = {
        status: 'success',
        data,
        code,
        order,
    };
    return (0, exports.complete)(partial);
};
exports.ok = ok;
const err = (failure, { message, code, order, skip } = {}) => {
    if (skip) {
        order = -Infinity;
    }
    let type;
    if ((0, checks_1.isErr)(failure)) {
        type = failure.type;
    }
    if (!type) {
        type = message || 'Unknown';
    }
    const exception = new FailureException(type, failure, order, message, code);
    if ((0, checks_1.isErr)(exception.error)) {
        if (exception.error.message === undefined) {
            delete exception.error.message;
        }
    }
    return (0, exports.complete)(exception);
};
exports.err = err;
const fail = (type, { message, code, order, skip, ...error } = {}) => {
    const failure = {
        ...error,
        type,
        message,
    };
    return (0, exports.err)(failure, { message, code, order, skip });
};
exports.fail = fail;
const throwFail = (type, params = {}) => {
    throw (0, exports.fail)(type, params);
};
exports.throwFail = throwFail;
const compare = (r1, r2) => {
    if ((r1.order || 0) > (r2.order || 0)) {
        return r1;
    }
    if ((r2.order || 0) > (r1.order || 0)) {
        return r2;
    }
    if (r1.isErr()) {
        return r1;
    }
    return r2;
};
exports.compare = compare;
const toResult = (input) => {
    if ((0, checks_1.isSuccessLike)(input)) {
        return (0, exports.ok)(input.data);
    }
    if ((0, checks_1.isFailureLike)(input)) {
        return (0, exports.err)(input.error);
    }
    throw new Error('Unexpected input');
};
exports.toResult = toResult;
function nope(_p) {
    throw new Error('Unreachable');
}
exports.nope = nope;
//# sourceMappingURL=utils.js.map